//Shared utilitary functions
//By Higor


class LCStatics expands Object;

#exec OBJ LOAD FILE="LCPureUtil.u" PACKAGE=LCWeapons_0002

const MULTIPLIER = 0x015a4e35;
const INCREMENT = 1;


//************************************************************************
//Use the util to get a player's rotation while keeping Pure compatibility
//************************************************************************
static function rotator PlayerRot( pawn P)
{
	return class'LCPureRotation'.static.PlayerRot(P);
}


//******************************************************************************************************
//Hash a player instead of referencing the actor, we're doing this because of security and BotZ support.
//******************************************************************************************************
static function int ffPCode( private Pawn ffOther)
{
	return ffOther.PlayerReplicationInfo.PlayerID;
/* BUGHERE
	local private PlayerReplicationInfo ffPRI;
	local private int ffTmp;
	local private int nani, nani2;

	ffPri = ffOther.PlayerReplicationInfo;
	ffTmp = ffPRI.PlayerID * 2;
	nani = ffTmp;
	ffSwap( nani, nani2);
	nani = ffRevertByte( nani2);
	ffSwap( nani, ffTmp);
	nani = ffPri.Deaths * 3 + 150;
	ffTmp = ffTmp | (nani + nani2);
	ffSwap( ffTmp, nani2);
	return nani2;*/
}

//*****************
//Swap two integers
//*****************
static function ffSwap( out private int U, out private int H)
{
	local private int L;
	
	L = U;
	U = H;
	H = L;
}


//*************************
//Client friendly TraceShot
//*************************
static function actor ffTraceShot(out vector ffHitLocation, out vector ffHitNormal, vector ffEndTrace, vector ffStartTrace, private Pawn ffTmp)
{
	local private vector ffRealHit;
	local private actor ffOther;

	ffOther = ffTmp.Trace(ffHitLocation,ffHitNormal,ffEndTrace,ffStartTrace,True);
	if ( Pawn(ffOther) != None )
	{
		ffRealHit = ffHitLocation;
		if ( !Pawn(ffOther).AdjustHitLocation(ffHitLocation, ffEndTrace - ffStartTrace) )
			ffOther = ffTraceShot(ffHitLocation,ffHitNormal,ffEndTrace,ffRealHit,Pawn(ffOther));
	}
	if ( ffOther == none )
		ffHitLocation = ffEndTrace;
	return ffOther;
}

//************************************
//Server TraceShot for missed ZP shots
//************************************
static function actor ffIrrelevantShot(out vector ffHitLocation, out vector ffHitNormal, vector ffEndTrace, vector ffStartTrace, private Pawn ffTmp, private float ffPing)
{
	local private vector ffRealHit;
	local private actor ffOther;

	if ( ffTmp == none )
		return none;
	ForEach ffTmp.TraceActors( class'Actor', ffOther, ffHitLocation, ffHitNormal, ffEndTrace, ffStartTrace)
	{
		if ( TraceStopper( ffOther) )
			return ffOther;
		if ( (!ffOther.bProjTarget && !ffOther.bBlockActors) || RelevantHitActor(ffOther, PlayerPawn(ffTmp), ffPing) ) //Non-solids or client hit actors ignored
			continue;
		if ( ffOther.bIsPawn && !Pawn(ffOther).AdjustHitLocation(ffHitLocation, ffEndTrace - ffStartTrace) )
			continue;
		return ffOther;
	}
	return none;
}


//*********************************
//Byte reversal in a simple integer
//*********************************
static function int ffRevertByte( private int nani2)
{
	local private int nani, HackMe;
	For ( HackMe=0 ; HackMe<32 ; HackMe++ )
		nani = nani | ( ((nani2 >>> HackMe) & 1) << (31-HackMe));
}


//********************************
//Turn a rotator into a single INT
//********************************
static function int CompressRotator( rotator R)
{
	return (R.Pitch << 16) | (R.Yaw & 65535);
}

//*************************
//Turn a INT into a rotator
//*************************
static function rotator DecompressRotator( int A)
{
	local rotator aRot;
	aRot.Yaw = A & 65535;
	aRot.Pitch = (A >>> 16);
	return aRot;
}


//*************************************************************
//Finds out if this hit actor should be considered for ZP shots
//*************************************************************
static function bool RelevantHitActor( actor Other, optional PlayerPawn P, optional float Ping)
{
	if ( Other.bIsPawn && (StationaryPawn(Other) == none) )
		return true;
	if ( Projectile(Other) != none )
	{
//		if ( sgWarShell(Other) != none && (Other.Level.TimeSeconds - sgWarShell(Other).SpawnedAt < 0.12) )
//			return false; //Killing it before seeing it on clients.
		if ( P != none )
		{
			if ( (P.Level.NetMode != NM_Client) && (Other.Instigator == P) ) //Player shoots his own stuff
			{
				if ( (ShockProj(Other) != none) && ((Other.Default.LifeSpan - Other.LifeSpan)/Other.Level.TimeDilation < Ping) )
					return false; //Combo possible on huge pings
			}
		}
		return true;
	}
	return false;
}

//******************************************************************
//This actor is always the end of a shot, regardless of other checks
//******************************************************************
static function bool TraceStopper( actor Other)
{
	if ( Other == Other.Level || Other.IsA('Mover') )
		return true;
	return false;
}


//****************************************************
//If this is a lag compensator, return the real victim
//****************************************************
static function actor CompensatedHitActor( actor Other, out vector HitLocation)
{
	if ( Other.bIsPawn || Other == Other.Level ) //Super fast checks
		return Other;
	if ( XC_LagCompensator(Other) != none )
	{
		HitLocation += XC_LagCompensator(Other).ffOwner.Location - Other.Location;
		return XC_LagCompensator(Other).ffOwner;
	}
	if ( XC_ComboTracker(Other) != none )
	{
		HitLocation += XC_ComboTracker(Other).MyProj.Location - Other.Location;
		return XC_ComboTracker(Other).MyProj;
	}
	return Other;
}

//*******************************************************************
//Parse a parameter from this command string using a custom delimiter
//*******************************************************************
static function string NextParameter( out string Commands, string Delimiter)
{
	local string result;
	local int i;
	
	if ( Delimiter == "" )
	{	result = Commands;
		Commands = "";
		return result;
	}

	i = InStr(Commands, Delimiter);
	if ( i < 0 )
	{
		result = Commands;
		Commands = "";
		return result;
	}
	if ( i == 0 ) //Idiot parse
	{
		Commands = Mid( Commands, Len(Delimiter));
		return NextParameter( Commands, Delimiter);
	}
	result = Left( Commands, i);
	Commands = Mid( Commands, i + Len(Delimiter) );
	return result;
}

//***********************************
//Remove initial spaces from a string
//***********************************
static function string ClearSpaces( string Text)
{
	local int i;

	i = InStr(Text, " ");
	while( i == 0 )
	{
		Text = Right(Text, Len(Text) - 1);
		i = InStr(Text, " ");
	}
	return Text;
}

//*************
//Replaces text
//*************
static function ReplaceText(out string Text, string Replace, string With)
{
	local int i;
	local string Input;
		
	Input = Text;
	Text = "";
	i = InStr(Input, Replace);
	while(i != -1)
	{	
		Text = Text $ Left(Input, i) $ With;
		Input = Mid(Input, i + Len(Replace));	
		i = InStr(Input, Replace);
	}
	Text = Text $ Input;
}

//**************************************
// Spawns an enhanced copy of a LCWeapon
//**************************************
//Customized to allow respawning with custom ammo amounts
function inventory SpawnCopy( pawn Other, Weapon W )
{
	local TournamentWeapon Copy;
	if( W.Level.Game.ShouldRespawn(self) )
	{
		Copy = W.spawn(Class,Other,,,rot(0,0,0));
		Copy.Tag           = W.Tag;
		Copy.Event         = W.Event;
		Copy.PickupAmmoCount = W.PickupAmmoCount;
		Copy.AmmoName		= W.AmmoName;
		if ( !bWeaponStay )
			GotoState('Sleeping');
	}
	else
		Copy = self;
	Copy.RespawnTime = 0.0;
	Copy.bHeldItem = true;
	Copy.bTossedOut = false;
	Copy.GiveTo( Other );
	Copy.Instigator = Other;
	Copy.GiveAmmo(Other);
	Copy.SetSwitchPriority(Other);
	if ( !Other.bNeverSwitchOnPickup )
		Copy.WeaponSet(Other);
	Copy.AmbientGlow = 0;
	return Copy;
}


//**********************************************************************
// Generates a semi random number based on a given seed, seed is updated
// Ranges from -1 to 1 *************************************************
static function float fRandom_Seed(float Scale, out int RandomSeed)
{
	local int aRs;
	local float Result;

	if ( Scale == 0 )
		Scale = 1;

	RandomSeed = MULTIPLIER * RandomSeed + INCREMENT;
	aRs = ((RandomSeed >>> 16) & 65535) - 32768; //Sign is kept, precision increased
//	Log("Seed is now: "$RandomSeed@" aRs is: "$aRs);
	Result = Scale * aRs / 32768f;
	return Result;
}

//*******************************************
// Provide quick aim error using random seeds
//*******************************************
static function vector StaticAimError( vector Y, vector Z, float Accuracy, int RandomSeed)
{
	return (fRandom_Seed( 500, RandomSeed) * Y + fRandom_Seed( 500, RandomSeed) * Z) * Accuracy;
}
static function vector SyncAimError( vector Y, vector Z, float Accuracy, int RandomSeed, out int NewSeed)
{
	NewSeed = RandomSeed;
	return (fRandom_Seed( 500, NewSeed) * Y + fRandom_Seed( 500, NewSeed) * Z) * Accuracy;
}

//*******************************
// Placement and comparison utils
//*******************************
static final function float HSize( vector aVec)
{
	return VSize(aVec * vect(1,1,0));
}

static final function bool ActorsTouching( actor A, actor B)
{
	if ( abs(A.Location.Z - B.Location.Z) > (A.CollisionHeight + B.CollisionHeight) )
		return false;
	return HSize( A.Location - B.Location) <= (A.CollisionRadius + B.CollisionRadius);
}

static final function vector VLerp( float Alpha, vector A, vector B)
{
	return A + (B-A) * Alpha;
}

static final function float GetAlpha( float Value, float A, float B)
{
	return fClamp( (Value-A) / (B-A) ,0,1);
}